{% for line in project.copyright_head.split('\n') %}{{ '# ' ~ line }}
{% endfor -%}
{% include 'root.pyproject.toml' %}

{% include 'bandit.pyproject.toml' %}

{% include 'black.pyproject.toml' %}

{% include 'coverage.pyproject.toml' %}

{% include 'flake8.pyproject.toml' %}

{% include 'isort.pyproject.toml' %}

{% include 'mypy.pyproject.toml' %}

{% include 'pyright.pyproject.toml' %}

{% include 'pytest.pyproject.toml' %}

{% include 'ruff.pyproject.toml' %}

{% include 'semantic_release.pyproject.toml' %}

{% include 'setuptools_scm.pyproject.toml' %}

{% include 'tox.pyproject.toml' %}
