{#-
# ozi/templates/tests/meson.build.j2
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
-#}
{{ '# OZI ' ~ spec.version }} meson.options
{%- for line in project.copyright_head.split('\n') %}
{{ '# ' ~ line }}
{%- endfor %}
{{ '# Additional copyright and license terms apply, see NOTICE for details.' }}
{{ '# DO NOT EDIT BY HAND - This file was generated for use with OZI tools.' }}
{{ '# ozi-fix -> meson rewrite is the standard editing pipeline.' }}
{% include 'project.array.meson.options' %}
{% include 'project.feature.meson.options' %}
{% include 'project.integer.meson.options' %}
{% include 'pyright.meson.options' %}
{% include 'mypy.meson.options' %}
{% include 'isort.meson.options' %}
{% include 'flake8.meson.options' %}
{% include 'black.meson.options' %}
{% include 'bandit.meson.options' %}
{% include 'pytest.meson.options' %}
