# CHANGELOG

## v0.0.0 ({{ spec.python.src.format.date|current_date }})

### :construction_worker:

* :construction_worker:: Initialized {{ project.name }} with ``ozi-new``.

```sh
ozi-new {{ project.argv }}
```
