{#- OZI 0.1
# ozi/templates/github_workflows/sigstore.yml
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
-#}
{%- raw %}
    - name: Sign wheel
{%- endraw %}
      if: steps.release.outputs.released == 'true' || steps.draft.outputs.created == 'true'
      run: |
{%- raw %}
        sigstore sign --output-directory sig dist/${{ github.event.repository.name }}-${{ steps.version.outputs.tag }}-${{ matrix.py }}-none-any.whl
{%- endraw %}
