{#- OZI 0.1
# ozi/templates/github_workflows/semantic_release.yml
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#}
    - name: Python Semantic Release
      id: release
      if: matrix.py == '{{ spec.python.support.security|wheel_repr }}'
      uses: python-semantic-release/python-semantic-release@v8.5.2
      with:
{%- raw %}
        github_token: ${{ secrets.GITHUB_TOKEN }}
{%- endraw %}
